import os

import requests
from authlib.integrations.base_client.errors import OAuthError
from dotenv import load_dotenv
from starlette import status

load_dotenv()

W3ID_USER_INFO_URL = os.getenv('W3ID_USER_INFO_URL')
W3ID_CLIENT_ID = os.getenv('W3ID_CLIENT_ID')
W3ID_CLIENT_SECRET = os.getenv('W3ID_CLIENT_SECRET')
W3ID_ENDPOINT_INTROSPECT = os.getenv('W3ID_ENDPOINT_INTROSPECT')
W3ID_ACCESS_TOKEN_URL = os.getenv('W3ID_ACCESS_TOKEN_URL')


# TODO: CSIAQ0158E The [authorization_grant] of type [authorization_code] does not exist or is invalid.
# def sso_ibm_get_access_token_by_code(code: str, redirect_uri: str):
#     '''Get access token by code.'''

#     headers = {
#         'Content-Type': 'application/x-www-form-urlencoded',
#     }

#     form_data = {
#         'code': code,
#         'grant_type': 'authorization_code',
#         'client_id': W3ID_CLIENT_ID,
#         'client_secret': W3ID_CLIENT_SECRET,
#         'redirect_uri': redirect_uri,
#     }
#     data = '&'.join([f'{key}={value}' for key, value in form_data.items()])

#     response = requests.post(
#         W3ID_ACCESS_TOKEN_URL,
#         headers=headers,
#         data=data,
#     )

#     result = response.json()
#     if (response.status_code == status.HTTP_200_OK):
#         return result
#     else:
#         raise OAuthError(error=result['error_description'])


def sso_ibm_refresh_access_token_by_refresh_token(refresh_token: str):
    '''Refresh access token by refresh token.'''

    headers = {
        'Content-Type': 'application/x-www-form-urlencoded',
    }

    form_data = {
        'refresh_token': refresh_token,
        'grant_type': 'refresh_token',
        'client_id': W3ID_CLIENT_ID,
        'client_secret': W3ID_CLIENT_SECRET,
    }
    data = '&'.join([f'{key}={value}' for key, value in form_data.items()])

    response = requests.post(
        W3ID_ACCESS_TOKEN_URL,
        headers=headers,
        data=data,
    )

    result = response.json()
    if (response.status_code == status.HTTP_200_OK):
        return result
    else:
        raise OAuthError(error=result['error'])


def sso_ibm_get_user_info(access_token: str):
    '''Get userinfo by access token'''

    headers = {
        'Content-Type': 'application/x-www-form-urlencoded',
    }

    form_data = {
        'access_token': access_token,
        'client_id': W3ID_CLIENT_ID,
        'client_secret': W3ID_CLIENT_SECRET,
    }
    data = '&'.join([f'{key}={value}' for key, value in form_data.items()])

    response = requests.post(
        W3ID_USER_INFO_URL,
        headers=headers,
        data=data,
    )

    result = response.json()
    if (response.status_code == status.HTTP_200_OK):
        return result
    else:
        raise OAuthError(error=result['error'])


def sso_ibm_verify_access_token(access_token: str):
    '''Verify access token'''

    # curl POST 'https://preprod.login.w3.ibm.com/oidc/endpoint/default/introspect' \
    # --header 'Content-Type: application/x-www-form-urlencoded' \
    # --data 'token=aaa&client_id=aaa&client_secret=aaa'

    headers = {
        'Content-Type': 'application/x-www-form-urlencoded',
    }

    form_data = {
        'token': access_token,
        'client_id': W3ID_CLIENT_ID,
        'client_secret': W3ID_CLIENT_SECRET,
    }
    data = '&'.join([f'{key}={value}' for key, value in form_data.items()])

    response = requests.post(
        W3ID_ENDPOINT_INTROSPECT,
        headers=headers,
        data=data,
    )

    result = response.json()
    if (response.status_code == status.HTTP_200_OK):
        if not result['active']:
            raise OAuthError(error=result['error'])
    else:
        raise OAuthError(error=result['error'])



#!/usr/bin/python3
# -*- coding:utf-8 -*-

# https://shiriev.ru/pydantic-two-way-mapping/
# https://docs.pydantic.dev/latest/migration/#changes-to-dataclasses
# https://docs.pydantic.dev/latest/concepts/dataclasses/#dataclass-config
# https://docs.pydantic.dev/latest/migration/#changes-to-config
# https://www.youtube.com/watch?app=desktop&v=Z0a0Vjd992I

from pydantic import BaseModel, Field
from pydantic.dataclasses import dataclass


@dataclass(config=dict(populate_by_name=True))
class UserInfoVO(BaseModel):
    display_name: str = Field(validation_alias='displayName')
    email_address: str = Field(validation_alias='emailAddress')
    uid: str


class TokenInfoVO(BaseModel):
    access_token: str
    refresh_token: str
    scope: str
    grant_id: str
    id_token: str
    token_type: str
    expires_in: int
    expires_at: int


@dataclass(config=dict(populate_by_name=True))
class AuthorizeInfoVO(BaseModel):
    access_token: str
    refresh_token: str
    scope: str
    grant_id: str
    id_token: str
    token_type: str
    expires_in: int
    expires_at: int
    user_info: UserInfoVO = Field(validation_alias='userinfo')


if __name__ == '__main__':
    data = {
        'displayName': 'displayName',
        'emailAddress': 'emailAddress',
        'uid': 'uid',
    }
    result = UserInfoVO.model_validate(data)
    print(result)



from setuptools import find_packages, setup


def read_files(files):
    data = []
    for file in files:
        with open(file, encoding='utf-8') as f:
            data.append(f.read())
    return "\n".join(data)


long_description = read_files(['README.md', 'CHANGELOG.md'])

meta = {}
with open('./src/ibm_sso/version.py', encoding='utf-8') as f:
    exec(f.read(), meta)

setup(
    name='ibm_sso',
    description='IBM SSO Self-Service Provisioner',
    long_description=long_description,
    long_description_content_type='text/markdown',
    version=meta['__version__'],
    author='xiaobitipao',
    author_email='xiaobitipao@gmail.com',
    url='https://github.com/xiaobitipao/ibm-sso',
    keywords=['sso', 'fastapi', 'oidc'],
    packages=find_packages('src'),    # include all packages under src
    package_dir={'': 'src'},          # tell distutils packages are under src
    python_requires='>=3.10',
    license='BSD-3-Clause',
    classifiers=[
        # https://pypi.org/classifiers/
        'Development Status :: 5 - Production/Stable',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.12',
        'Programming Language :: Python :: Implementation :: PyPy',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
    ]
)



import os

from authlib.integrations.base_client.errors import OAuthError
from authlib.integrations.requests_client import OAuth2Session
from authlib.integrations.starlette_client import OAuth
from dotenv import load_dotenv
from fastapi import APIRouter, Depends, Query, Security
from fastapi.security import HTTPAuthorizationCredentials, HTTPBearer
from starlette.config import Config
from starlette.requests import Request
from starlette.responses import HTMLResponse, RedirectResponse

from ibm_sso.service.sso_ibm_service import sso_ibm_get_user_info
from ibm_sso.vo.UserInfoVO import AuthorizeInfoVO, TokenInfoVO, UserInfoVO

authorize_router = APIRouter()

load_dotenv()

W3ID_CLIENT_ID = os.getenv('W3ID_CLIENT_ID')
W3ID_CLIENT_SECRET = os.getenv('W3ID_CLIENT_SECRET')
W3ID_ACCESS_TOKEN_URL = os.getenv('W3ID_ACCESS_TOKEN_URL')
W3ID_ENDPOINT_DISCOVERY = os.getenv('W3ID_ENDPOINT_DISCOVERY')
W3ID_ENDPOINT_REVOCATION = os.getenv('W3ID_ENDPOINT_REVOCATION')

# Get environment variable information by default
config = Config()

# Create an authentication server instance based on the configuration
#
# from authlib.integrations.starlette_client.apps import StarletteOAuth2App
# class: authlib.integrations.starlette_client.apps.StarletteOAuth2App
oauth = OAuth(config)
w3id = oauth.register(
    name='w3id',
    server_metadata_url=W3ID_ENDPOINT_DISCOVERY,
    client_kwargs={
        'scope': 'openid profile email'
    }
)
# print(type(w3id))

# Create OAuth2Session instance
oauth2_session = OAuth2Session(
    W3ID_CLIENT_ID,
    W3ID_CLIENT_SECRET,
    scope='openid',
)


# #########################################################################################
# Swagger UI: Use the user input(username/password) to perform login processing through the endpoint(/token)
# from fastapi.security import OAuth2PasswordBearer
# oauth2_scheme = OAuth2PasswordBearer(tokenUrl='token')

# async def get_current_user(token: str = Depends(oauth2_scheme)):
#     '''Protect RESTful API'''
#     # sso_ibm_verify_access_token(token)
#     # return token
#     user_info = sso_ibm_get_user_info(token)
#     return user_info

# #########################################################################################
# Swagger UI: Use the user input(Bearer Token)
security = HTTPBearer()


async def get_current_user(credentials: HTTPAuthorizationCredentials = Security(security)):
    '''Protect RESTful API'''
    result = sso_ibm_get_user_info(credentials.credentials)
    return UserInfoVO.model_validate(result)


# TODO: CSIAQ0158E The [authorization_grant] of type [authorization_code] does not exist or is invalid.
# @authorize_router.post('/token', description='Get an access token by code.', summary='Obtain an Access Token', response_model=TokenInfoVO)
# async def get_token(dto: TokenDTO):
#     result = sso_ibm_get_access_token_by_code(dto.code, dto.redirect_uri)
#     return TokenInfoVO.model_validate(result)


@authorize_router.post(
    '/revoke',
    description='''Revoke access_token or refresh_token.<br/><br/>
        👇<br/>
        <strong><i>
        If the access token is revoked, the refresh token will not be revoked and can continue to be used.<br/>
        If the refresh token is revoked, the access token will also be revoked at the same time.
        </i></strong>
        ''',
    summary='Revoke access_token or refresh_token'
)
async def revoke_token(token: str):
    result = oauth2_session.revoke_token(
        W3ID_ENDPOINT_REVOCATION,
        # token_type_hint=token_type_hint.value,
        token=token,
    )
    return result.reason


@authorize_router.post(
    '/refresh_token',
    description='Get an new access token by refresh_token.',
    summary='Refresh the Access Token',
    response_model=TokenInfoVO
)
async def refresh_token(refresh_token: str):
    # result = sso_ibm_refresh_access_token_by_refresh_token(refresh_token)
    result = oauth2_session.refresh_token(
        W3ID_ACCESS_TOKEN_URL,
        refresh_token=refresh_token,
    )
    return TokenInfoVO.model_validate(result)


@authorize_router.get(
    '/userinfo',
    description='''Get user info.<br/><br/>
        👇<br/>
        <strong><i>
        For sso authentication, you can directly click 
        <a href='https://localhost:5000/oauth2/login?redirect=https://localhost:3000/example' target='_blank'>Login with SSO</a>
        </i></strong>
        ''',
    summary='Get user info',
    response_model=UserInfoVO
)
async def get_userinfo(user_info: UserInfoVO = Depends(get_current_user)):
    # curl -k GET 'https://localhost:5000/oauth2/userinfo' \
    # --header 'Authorization: Bearer access_token'
    # result = sso_ibm_get_user_info(token)
    return user_info


@authorize_router.get('/login', include_in_schema=False)
async def login(request: Request, redirect: str = Query(...)):

    request.session['redirect'] = redirect

    redirect_uri = request.url_for('auth')
    return await w3id.authorize_redirect(request, redirect_uri)


@authorize_router.get('/authorize', include_in_schema=False)
async def auth(request: Request):
    try:
        token = await w3id.authorize_access_token(request)
    except OAuthError as error:
        # TODO: Error page
        return HTMLResponse(f'<h1>{error.error}</h1>')

    user = token.get('userinfo')
    if user:
        request.session['user'] = dict(user)

    # param = json.dumps(token)
    token_info = AuthorizeInfoVO.model_validate(token)
    redirect = request.session['redirect']
    return RedirectResponse(url=f'{redirect}?response={token_info.model_dump_json()}')
